==> fstringen-0.0.2-py3-none-any.whl: fstringen/fstringen.py <==
import inspect
import atexit
import json
import re
import textwrap

try:
    import yaml
except ImportError:
    yaml = None


def _putify(template):
    lines = template.split("\n")
    newlines = []
    for line in lines:
        # We don't care about indent in single-line strings
        if len(lines) == 1:
            indent = ''
        else:
            indent = line[:len(line) - len(line.lstrip())]
        # Take out all the isolated '{' and '}' characters
        line = line.replace("{{", "#_fstringen_ocbblock#").replace(
            "}}", "#_fstringen_ccbblock#")
        nl = line.replace("{", "{_put(").replace("}", ", '" + indent + "')}")
        # Put back all the isolated '{' and '}' characters
        nl = nl.replace("#_fstringen_ocbblock#",
                        "{{").replace("#_fstringen_ccbblock#", "}}")
        newlines.append(nl)
    return "\n".join(newlines)


def _put(obj, indent):
    if type(obj) in (list, tuple):
        newobj = [str(el).replace("\n", "\n" + indent) for el in obj]
        return ("\n" + indent).join(newobj)
    else:
        return str(obj).replace("\n", "\n" + indent)


def _fix_fstring(line):
    line = re.sub(r"(\"\"\"\*\s*)", "\"\"\"", line)
    line = re.sub(r"(\s*\*\"\"\")", "\"\"\"", line)
    return line


def _compile(fstringstar):
    globals_ = inspect.currentframe().f_back.f_globals
    locals_ = inspect.currentframe().f_back.f_locals

    fstring = "_fstringstar = f\"\"\"{}\"\"\"".format(_putify(fstringstar))

    exec(fstring, globals_, locals_)
    output = locals_["_fstringstar"]
    return output


def gen(model=None, fname=None, comment=None, notice=True):
    def realgen(fn):
        original_name = fn.__name__
        code = textwrap.dedent(inspect.getsource(fn))
        newcode = code[1:]  # Remove decorator
        newcode = re.sub(r"(f\"\"\"\*)", "_compile(\"\"\"", newcode)
        newcode = re.sub(r"(\*\"\"\")", "\"\"\")", newcode)

        # Re-execute function definition with the new code, in its own
        # globals/locals scope
        globals_ = inspect.currentframe().f_back.f_globals
        globals_["_put"] = _put
        globals_["_compile"] = _compile
        locals_ = inspect.currentframe().f_back.f_locals
        exec(newcode, globals_, locals_)
        newgen = locals_[original_name]

        def newfn(*args, **kwargs):
            r = newgen(*args, **kwargs)
            v = textwrap.dedent(r)
            if len(v) > 0 and v[0] == "\n":
                v = v[1:]
            if len(v) > 0 and v[-1] == "\n":
                v = v[:-1]
            return v

        if fname:
            if notice and not comment:
                raise FStringenError(
                    "Cannot place a notice without having the comment prefix")
            global _output
            _output[fname] = {
                "notice": notice,
                "comment": comment,
                "fn": newfn,
                "model": model
            }

        # Put the new function in globals, so others @gen code can call it
        # This is obviously dangerous, and it's one of the the reasons why
        # fstringen must not be used in anything else other than generators.
        globals_[original_name] = newfn

        return newfn

    return realgen


_output = {}


def _generate_all():
    for fname in _output:
        genopts = _output[fname]
        f = open(fname, "w")
        if genopts["notice"]:
            f.write(genopts["comment"] +
                    " File generated by fstringen. DO NOT EDIT.\n\n")
        fn = genopts["fn"]
        model = genopts["model"]
        f.write(fn(model))
        f.close()


atexit.register(_generate_all)


class FStringenError(Exception):
    pass


class FStringenNotFound(FStringenError):
    pass


class Mapper:
    def __init__(self, name, mappings):
        self.name = name
        self.mappings = mappings

    def __getitem__(self, key):
        return self.get(key)

    def get(self, key):
        return self.mappings.get(
            key, "<! MAPPING NOT FOUND IN '{}': {} !>".format(self.name, key))


class Model:
    def __init__(self, name, model, refindicator="#", root_model=None):
        self.name = name
        self.model = model
        self.refindicator = refindicator
        self.root_model = root_model
        if self.root_model is None:
            self.root_model = model

    @staticmethod
    def fromYAML(fname, refindicator="#", root_model=None):
        if yaml is None:
            raise FStringenError("Cannot find 'yaml' module")
        return Model(fname,
                     yaml.load(open(fname, "r").read(), Loader=yaml.Loader),
                     refindicator, root_model)

    @staticmethod
    def fromJSON(fname, refindicator="#", root_model=None):
        return Model(fname, json.loads(open(fname, "r").read()), refindicator,
                     root_model)

    @staticmethod
    def fromDict(dict_, name="dict", refindicator="#", root_model=None):
        return Model(name, dict_, refindicator, root_model)

    def __iter__(self):
        return iter(self.model)

    def __getitem__(self, key):
        if type(self.model) in (list, tuple):
            try:
                key = int(key)
            except ValueError:
                raise FStringenError(
                    "array model navigation requires integers")
        return self.model[key]

    def __str__(self):
        return str(self.model)

    def __repr__(self):
        return repr(self.model)

    def dict(self):
        return self.model

    def __eq__(self, other):
        return self.model == other.model and \
            self.name == other.name and \
            self.refindicator == other.refindicator and \
            self.root_model == other.root_model

    def has(self, path):
        callerctx = inspect.currentframe().f_back
        try:
            self.select(path, callerctx)
        except FStringenNotFound:
            return False

        return True

    def _new(self, name, submodel):
        return Model(name, submodel, self.refindicator, self.root_model)

    def select(self, path, callerctx=None):
        if callerctx is None:
            callerctx = inspect.currentframe().f_back

        obj = self.model
        root = False
        # Ignore ref indicators and browse accordingly
        if path.startswith(self.refindicator):
            path = path[1:]
        # When an absolute path is used in a query, revert to the root model
        if path.startswith("/"):
            path = path[1:]
            obj = self.root_model
            root = True
        # Empty path trailings are ignored
        if path.endswith("/"):
            path = path[:-1]

        parts = path.split("/")
        curpath = []
        if root:
            curpath.append("")

        lastpart = None

        # TODO: we should probably browse models recursively instead of in a
        # loop
        for i in range(len(parts)):
            part = parts[i]

            match = re.match(r"\<([^<>]*)\>", part)
            if match is not None and match.lastindex == 1:
                var = match[1]
                if not var.isidentifier():
                    raise FStringenError(
                        "'{}' is not a valid Python identifier".format(var))
                if var in callerctx.f_locals:
                    part = str(callerctx.f_locals[var])
                    curpath.append("<{}>".format(part))
                else:
                    raise FStringenError(
                        "Could not find '{}' in local scope".format(var))
                if part.startswith(self.refindicator):
                    return self.select(part[1:], callerctx)
            else:
                curpath.append(part)

            if part == "*" and i == len(parts) - 1:
                if not hasattr(obj, "items"):
                    raise FStringenError("cannot iterate over '{}'".format(
                        "/".join(curpath[:-1])))
                return tuple(self._new(k, v) for k, v in obj.items())
            elif part.endswith("->"):
                part = part[:-2]
                newpath = obj[part]
                newmodel = self._new(part, obj)
                return newmodel.select(newpath, callerctx)
            else:
                if type(obj) in (list, tuple):
                    try:
                        part = int(part)
                    except ValueError:
                        raise FStringenError(
                            "array model navigation requires integers")
                try:
                    obj = obj[part]
                except (KeyError, IndexError):
                    raise FStringenNotFound(
                        "could not find path '{}' in model '{}'".format(
                            "/".join(curpath), obj))

            lastpart = part

        return self._new(lastpart, obj) if type(obj) in (list, tuple,
                                                         dict) else obj

==> fstringen-0.0.4-py3-none-any.whl: fstringen/fstringen.py <==
import inspect
import atexit
import json
import re
import textwrap

try:
    import yaml
except ImportError:
    yaml = None


def _putify(template):
    lines = template.split("\n")
    newlines = []
    for line in lines:
        # We don't care about indent in single-line strings
        if len(lines) == 1:
            indent = ''
        else:
            indent = line[:len(line) - len(line.lstrip())]
        # Take out all the isolated '{' and '}' characters
        line = line.replace("{{", "#_fstringen_ocbblock#").replace(
            "}}", "#_fstringen_ccbblock#")
        nl = line.replace("{", "{_put(").replace("}", ", '" + indent + "')}")
        # Put back all the isolated '{' and '}' characters
        nl = nl.replace("#_fstringen_ocbblock#",
                        "{{").replace("#_fstringen_ccbblock#", "}}")
        newlines.append(nl)
    return "\n".join(newlines)


def _put(obj, indent):
    if isinstance(obj, list) or isinstance(obj, tuple):
        newobj = [str(el).replace("\n", "\n" + indent)
                  for el in obj if el is not None]
        return ("\n" + indent).join(newobj)
    else:
        if obj is None:
            return ""
        return str(obj).replace("\n", "\n" + indent)


def _fix_fstring(line):
    line = re.sub(r"(\"\"\"\*\s*)", "\"\"\"", line)
    line = re.sub(r"(\s*\*\"\"\")", "\"\"\"", line)
    return line


def _compile(fstringstar):
    globals_ = inspect.currentframe().f_back.f_globals
    locals_ = inspect.currentframe().f_back.f_locals

    fstring = "_fstringstar = f\"\"\"{}\"\"\"".format(_putify(fstringstar))

    exec(fstring, globals_, locals_)

    output = locals_["_fstringstar"]
    return output


def gen(model=None, fname=None, comment=None, notice=True):
    def realgen(fn):
        original_name = fn.__name__
        code = textwrap.dedent(inspect.getsource(fn))
        newcode = code[1:]  # Remove decorator
        newcode = re.sub(r"(f\"\"\"\*)", "_compile(\"\"\"", newcode)
        newcode = re.sub(r"(\*\"\"\")", "\"\"\")", newcode)

        # Re-execute function definition with the new code, in its own
        # globals/locals scope
        globals_ = inspect.currentframe().f_back.f_globals
        globals_["_put"] = _put
        globals_["_compile"] = _compile
        locals_ = inspect.currentframe().f_back.f_locals
        exec(newcode, globals_, locals_)
        newgen = locals_[original_name]

        def newfn(*args, **kwargs):
            r = newgen(*args, **kwargs)
            if r is None:
                return
            v = textwrap.dedent(r)
            if len(v) > 0 and v[0] == "\n":
                v = v[1:]
            if len(v) > 0 and v[-1] == "\n":
                v = v[:-1]
            return v

        if fname:
            if notice and not comment:
                raise FStringenError(
                    "Cannot place a notice without having the comment prefix")
            global _output
            _output[fname] = {
                "notice": notice,
                "comment": comment,
                "fn": newfn,
                "model": model
            }

        # Put the new function in globals, so others @gen code can call it
        # This is obviously dangerous, and it's one of the the reasons why
        # fstringen must not be used in anything else other than generators.
        globals_[original_name] = newfn

        return newfn

    return realgen


_output = {}


def _generate_all():
    for fname in _output:
        genopts = _output[fname]
        f = open(fname, "w")
        if genopts["notice"]:
            f.write(genopts["comment"] +
                    " File generated by fstringen. DO NOT EDIT.\n\n")
        fn = genopts["fn"]
        model = genopts["model"]
        f.write(fn(model))
        f.close()


atexit.register(_generate_all)


class FStringenError(Exception):
    pass


class FStringenNotFound(FStringenError):
    pass


class Mapper:
    def __init__(self, name, mappings):
        self.name = name
        self.mappings = mappings

    def __getitem__(self, key):
        return self.get(key)

    def get(self, key):
        return self.mappings.get(
            key, "<! MAPPING NOT FOUND IN '{}': {} !>".format(self.name, key))


class Selectable:
    def __new__(cls, name, value, refindicator="#", root=None):
        # Pick Selectable methods that should be used
        methods = {}
        for method in cls.__dict__:
            if not method.startswith("__"):
                methods[method] = cls.__dict__[method]

        # Python does not allow subclassing bool, so we use an adapted int
        if type(value) == bool:
            value = int(value)

            def custom_repr_str(instance):
                return str(bool(instance.value))

            methods["__repr__"] = custom_repr_str
            methods["__str__"] = custom_repr_str
        # None cannot be subclassed either, use an empty string instead
        elif value is None:
            value = ""

            def custom_repr_str(instance):
                return "None"

            def custom_eq(instance, other):
                return other is None

            methods["__repr__"] = custom_repr_str
            methods["__str__"] = custom_repr_str
            methods["__eq__"] = custom_eq

        # New type will be a subclass of the value, with the extra methods
        # defined in Selectable
        newcls = type(cls.__name__, (type(value),), methods)
        obj = newcls(value)
        # Initialize non-value attributes
        obj._initSelectable(name, refindicator, root)
        return obj

    def _initSelectable(self, name, refindicator, root):
        self.name = name
        self.value = self
        self.refindicator = refindicator
        self.root = root
        if self.root is None:
            self.root = self.value

    @staticmethod
    def fromYAML(fname, refindicator="#", root=None):
        """ Loads a Selectable from a YAML file """
        if yaml is None:
            raise FStringenError("Cannot find 'yaml' module")
        return Selectable(fname,
                     yaml.load(open(fname, "r").read(), Loader=yaml.Loader),
                     refindicator, root)

    @staticmethod
    def fromJSON(fname, refindicator="#", root=None):
        """ Loads a Selectable from a JSON file """
        return Selectable(fname, json.loads(open(fname, "r").read()), refindicator,
                     root)

    @staticmethod
    def fromDict(dict_, name="dict", refindicator="#", root=None):
        """ Loads a Selectable from a Python dictionary """
        return Selectable(name, dict_, refindicator, root)

    def has(self, path):
        """ Returns True if path exists in the Selectable """
        callerctx = inspect.currentframe().f_back
        try:
            self.select(path, callerctx)
        except FStringenNotFound:
            return False

        return True

    def isreference(self, path):
        """ Returns True if path is a possible reference """
        callerctx = inspect.currentframe().f_back
        value = self.select(path, callerctx)
        return isinstance(value, str) and value.startswith(self.refindicator)

    def _new(self, name, selectable):
        """ Instantiates a Selectable keeping the same root """
        return Selectable(name, selectable, self.refindicator, self.root)

    def select(self, path, callerctx=None):
        """ Returns a new Selectable based on path """

        if callerctx is None:
            callerctx = inspect.currentframe().f_back

        obj = self.value
        root = False
        # Ignore ref indicators and browse accordingly
        if path.startswith(self.refindicator):
            path = path[1:]
        # When an absolute path is used in a query, revert to the root
        if path.startswith("/"):
            path = path[1:]
            obj = self.root
            root = True
        # Empty path trailings are ignored
        if path.endswith("/"):
            path = path[:-1]

        parts = path.split("/")
        curpath = []
        if root:
            curpath.append("")

        lastpart = None

        for i in range(len(parts)):
            part = parts[i]

            # Replace <variables> with their values
            match = re.match(r"\<([^<>]*)\>", part)
            if match is not None and match.lastindex == 1:
                var = match[1]
                if not var.isidentifier():
                    raise FStringenError(
                        "'{}' is not a valid Python identifier".format(var))
                if var in callerctx.f_locals:
                    part = str(callerctx.f_locals[var])
                    curpath.append("<{}>".format(part))
                else:
                    raise FStringenError(
                        "Could not find '{}' in local scope".format(var))
                if part.startswith(self.refindicator):
                    return self.select(part[1:], callerctx)
            else:
                curpath.append(part)

            if part == "*" and i == len(parts) - 1:
                if not hasattr(obj, "items"):
                    raise FStringenError("cannot iterate over '{}'".format(
                        "/".join(curpath[:-1])))
                elements = tuple(self._new(k, v) for k, v in obj.items())
                return self._new("*", elements)
            elif part.endswith("->"):
                part = part[:-2]
                newpath = obj[part]
                newselectable = self._new(part, obj)
                return newselectable.select(newpath, callerctx)
            else:
                if isinstance(obj, list) or isinstance(obj, tuple):
                    try:
                        part = int(part)
                    except ValueError:
                        raise FStringenError(
                            "array navigation requires integers")
                elif not isinstance(obj, dict):
                    raise FStringenError(
                            "cannot lookup path '{}' in value '{}'".format(
                                part, str(obj)))
                try:
                    obj = obj[part]
                except (KeyError, IndexError):
                    raise FStringenNotFound(
                        "could not find path '{}' in '{}'".format(
                            "/".join(curpath), obj))

            lastpart = part

        return self._new(lastpart, obj)


Model = Selectable
